#!/usr/bin/python
import mdreader
import numpy as np
from scipy.optimize import curve_fit
import matplotlib as mpl
mpl.use('Agg')
import matplotlib.pyplot as plt
from matplotlib.ticker import MultipleLocator, FormatStrFormatter


def getCenter(base, sc1, sc2):
    ''' This function takes as input the type of base and 
    coordinates of three first SC beads. It takes a linear 
    combination of these coordinates to produce the coordinates 
    of the center of dsDNA. The coefficients have been 
    optimized for each base separately using the ideal dsDNA
    structure. 
    '''
    if base == 'DA':
        coefs = np.array([0.49997824, 0.50001173])
    elif base == 'DC':
        coefs = np.array([0.49998997, 0.49998997])
    elif base == 'DG':
        coefs = np.array([0.49995035, 0.50002962])
    else:
        coefs = np.array([0.50010777, 0.49988221])
    return (  coefs[0]*np.array(sc1) 
            + coefs[1]*np.array(sc2)) 


def getAverageLength(coords1, coords2):
    ''' This gets the timeseries of two coordinates and calculates the vector 
    between them at each frame. Then it calculates the length of of the vector
    for each frame and normalizes that. Finally it calculates the average 
    length (in time) and returns that with the normalized vector.
    '''
    vec = coords2-coords1
    len = np.sqrt((vec*vec).sum(axis=1))
    normalized = vec/len[:,None]
    avg = np.average(len)
    return avg, normalized


def expfunc(x, lps):
    ''' Exponential function with two free parameters.
    '''
    return np.exp(-x*avg_lengths[0]/lps)


def writeCenterCoords(centers):
    ''' Saves the coordinates of the centers to a gro file.
    '''
    grofile = open('center-coordinates.gro', 'w')
    grofile.write('Coordinates for DNA centers\n')
    grofile.write(' {}\n'.format(centers.shape[1]))
    for i in range(centers.shape[1]):
        grofile.write('    1CE     CEN    1   {:1.3f}   {:1.3f}   {:1.3f}\n'.format(
            centers[0,i,0], centers[0,i,1], centers[0,i,2]))
    grofile.write('   100.000 100.000 100.000\n')
    grofile.close()


def initializeMDreader():
    ''' Initializes a new MDreader object, adds an argument for the needed
    index group and reads the trajectory. In this case the index file should 
    contain only the SC1 beads from one strand for calculation of strand
    centers. Coordinates of the time series are then converted from angstroms 
    '''
    md = mdreader.MDreader()
    md.add_argument("-oplot", dest="oplot",type=str)
    md.add_ndx(ng=1, ndxparms=["Select atoms"])
    md.do_parse()
    ndx_group = md.ndxgs[0]
    tseries1 = md.timeseries(md.ndxgs[0])
    coords = tseries1.coords/10
    return ndx_group, coords, md.opts.oplot


def calcCenters(ndx_group, coords):
    ''' The center of each bp is calculated from the SC1 bead coordinates.
    This is done using coefficients optimized with an ideal dsDNA structure. 
    Coordinates of centers are returned.
    '''
    centers = np.zeros([coords.shape[0],coords.shape[1]/2,coords.shape[2]])
    for i in range(centers.shape[1]):
        centers[:,i,:] = getCenter(ndx_group.resnames[i], 
                                   coords[:,i,:], 
                                   coords[:,-i-1,:])
    return centers


def calcPersistence(indices, dot_products):
    ''' Fits and exponential decay to the dot products and calculates 
    persistence length based on that from equation
    <n0.ni> = exp(-i*<l0>/lp), where lp is persistence length. It then 
    prints this out in nanometers and basepairs (based on the length of 
    first vector) and returns both. 
    '''
    popt, pvoc = curve_fit(expfunc, indices, dot_products)
    persistence_nm = popt[0]
    persistence_bp = popt[0]/avg_lengths[0]
    return persistence_nm, persistence_bp


def writeData(indices, dot_products, persistence_nm, stds):
    fit = expfunc(indices, persistence_nm)
    np.save('fit.npy', fit)
    np.save('dot-products.npy', dot_products)
    np.save('indices.npy', indices)
    np.save('stds.npy', stds)


def plotDecay(indices, dot_products, persistence_nm, stds, length, output):
    fig, ax = plt.subplots(1, 1)
    fig.set_size_inches(6,3)
    ax.errorbar(indices, dot_products, yerr=stds, fmt='o', color='b')
    ax.plot(indices, expfunc(indices, persistence_nm), linewidth=2, color='b')
    ax.set_xlim([0,length])
    ax.set_yscale('log')
    ax.set_ylim([0.6,1.0])
    # Set ticks
    ax.xaxis.set_ticks_position('bottom')
    ax.yaxis.set_ticks_position('left')
    ax.tick_params(direction='out', length=4, width=1, which='minor')
    ax.tick_params(direction='out', length=5, width=1, which='major')
    majorLocator   = MultipleLocator(20)
    majorFormatter = FormatStrFormatter('%d')
    minorLocator   = MultipleLocator(10)
    ax.xaxis.set_major_locator(majorLocator)
    ax.xaxis.set_major_formatter(majorFormatter)
    ax.xaxis.set_minor_locator(minorLocator)
    majorLocator   = MultipleLocator(0.10)
    majorFormatter = FormatStrFormatter('%.1f')
    minorLocator   = MultipleLocator(0.02)
    ax.yaxis.set_major_locator(majorLocator)
    ax.yaxis.set_major_formatter(majorFormatter)
    ax.yaxis.set_minor_locator(minorLocator)
    # Save figure
    fig.savefig('{}.png'.format(output), dpi=300, transparent='True')
    fig.savefig('{}.eps'.format(output), transparent='True')


# Initialize
ndx_group, coords, oplot = initializeMDreader()
centers = calcCenters(ndx_group, coords)

# Calculate normalized vectors between two consecutive centers 
# and their average lengths. 
ignore = 2
avg_lengths = np.zeros([centers.shape[1]-1-2*ignore])
normalized_vectors = np.zeros([centers.shape[0],
                               centers.shape[1]-1-2*ignore,
                               centers.shape[2]])
for i in range(avg_lengths.shape[0]):
    avg_lengths[i], normalized_vectors[:,i,:] = getAverageLength(
        centers[:,i+ignore,:], centers[:,i+ignore+1,:])

# Calculate the average dot products of each other vector with the first one.
slices = 5
length = normalized_vectors.shape[1]-15
dot_products = np.zeros([length-1, normalized_vectors.shape[1]+1-length])
slice_dot_products = np.zeros([slices,length-1, normalized_vectors.shape[1]+1-length])
slice_averages = np.zeros([slices, dot_products.shape[0]])
slice_stds = np.zeros([slices, dot_products.shape[0]])
averages = np.zeros(dot_products.shape[0])
stds = np.zeros(dot_products.shape[0])
# Analyze the data in slices. 
slice_ = int(np.floor(normalized_vectors.shape[0]/float(slices)))
for i in range(dot_products.shape[0]):
    for j in range(dot_products.shape[1]):
        #print i+j+1
        dot_products[i,j] = np.average(
            np.sum(normalized_vectors[:,j,:] * normalized_vectors[:,j+i+1,:], axis=1))
        for k in range(slices):
            slice_dot_products[k,i,j] = np.average(
                np.sum(normalized_vectors[k*slice_:(k+1)*slice_,j,:]
                       * normalized_vectors[k*slice_:(k+1)*slice_,j+i+1,:], axis=1))
    averages[i] = np.average(dot_products[i,:])
    stds[i] = np.std(dot_products[i,:])
    for k in range(slices):
        slice_averages[k,i] = np.average(slice_dot_products[k,i,:])

# Create a list for vector indices, calculate persistence length and save stuff
indices = np.array(range(1, dot_products.shape[0]+1))
persistence_nm, persistence_bp = calcPersistence(indices, averages)
timeslice_nm = np.zeros(slices)
timeslice_bp = np.zeros(slices)
for k in range(slices):
    timeslice_nm[k], timeslice_bp[k] = calcPersistence(indices, slice_averages[k])
print '\n\nPersistence length in nm and bp:'
print '{:.1f} +/- {:.1f} nm     {:.1f} +/- {:.1f} bp'.format(np.average(timeslice_nm), np.std(timeslice_nm), np.average(timeslice_bp), np.std(timeslice_bp))
print 'When averaging over {} slices.'.format(slices)
slice_nm = np.zeros(dot_products.shape[1])
slice_bp = np.zeros(dot_products.shape[1])
for i in range(dot_products.shape[1]):
    slice_nm[i], slice_bp[i] = calcPersistence(indices, dot_products[:,i])
print '{:.1f} +/- {:.1f} nm     {:.1f} +/- {:.1f} bp'.format(persistence_nm, np.std(slice_nm), persistence_bp, np.std(slice_bp))
print 'When averaging over different starting points.'
if oplot != '':
    plotDecay(indices, averages, persistence_nm, stds, length, 'vector-decay-{}'.format(oplot))
else:
    plotDecay(indices, averages, persistence_nm, stds, length, 'vector-decay')
#writeCenterCoords(centers)
#writeData(indices, averages, persistence_nm, stds)



